@@ -1,5 +1,10 @@
 #include "vconsole.h"
 
+static GtkWidget *tab_label_with_close_button(const char *labeltext,
+                                              GCallback callback,
+                                              gpointer opaque);
+static void domain_close_tab_btn(GtkWidget *btn, gpointer opaque);
+
 /* ------------------------------------------------------------------ */
 
 static const char *state_name[] = {
@@ -10,9 +15,7 @@ static const char *state_name[] = {
     [ VIR_DOMAIN_SHUTDOWN ]    = "shutdown",
     [ VIR_DOMAIN_SHUTOFF ]     = "shutoff",
     [ VIR_DOMAIN_CRASHED ]     = "crashed",
-#ifdef VIR_DOMAIN_PMSUSPENDED
     [ VIR_DOMAIN_PMSUSPENDED ] = "suspended",
-#endif
 };
 
 static const char *domain_state_name(struct vconsole_domain *dom)
@@ -68,21 +71,23 @@ static void domain_configure_vte(struct vconsole_domain *dom)
 {
     struct vconsole_window *win = dom->conn->win;
     VteTerminal *vte = VTE_TERMINAL(dom->vte);
-    VteTerminalCursorBlinkMode bl =
+    VteCursorBlinkMode bl =
         win->tty_blink ? VTE_CURSOR_BLINK_ON : VTE_CURSOR_BLINK_OFF;
-    GdkColor fg = {0,0,0,0};
-    GdkColor bg = {0,0,0,0};
+    GdkRGBA fg = { 0, 0, 0, 0 };
+    GdkRGBA bg = { 0, 0, 0, 0 };
+    PangoFontDescription *font;
 
     if (!dom->vte)
         return;
 
-    gdk_color_parse(win->tty_fg, &fg);
-    gdk_color_parse(win->tty_bg, &bg);
-
-    vte_terminal_set_font_from_string(vte, win->tty_font);
+    gdk_rgba_parse(&fg, win->tty_fg);
+    gdk_rgba_parse(&bg, win->tty_bg);
     vte_terminal_set_cursor_blink_mode(vte, bl);
     vte_terminal_set_color_foreground(vte, &fg);
     vte_terminal_set_color_background(vte, &bg);
+
+    font = pango_font_description_from_string(win->tty_font);
+    vte_terminal_set_font(vte, font);
 }
 
 static void make_dirs(const char *filename)
@@ -324,6 +329,7 @@ static void domain_update_tree_store(struct vconsole_domain *dom,
     const char *foreground;
     char load[16], mem[16];
     PangoWeight weight;
+    int avgload;
 
     switch (dom->info.state) {
     case VIR_DOMAIN_RUNNING:
@@ -337,6 +343,7 @@ static void domain_update_tree_store(struct vconsole_domain *dom,
     }
     snprintf(load, sizeof(load), "%d%%", dom->load);
     snprintf(mem, sizeof(mem), "%ld M", dom->info.memory / 1024);
+    avgload = dom->info.nrVirtCpu ? dom->load / dom->info.nrVirtCpu : 0;
 
     gtk_tree_store_set(dom->conn->win->store, guest,
                        NAME_COL,       dom->name,
@@ -346,7 +353,7 @@ static void domain_update_tree_store(struct vconsole_domain *dom,
                        STATE_COL,      domain_state_name(dom),
                        NR_CPUS_COL,    dom->info.nrVirtCpu,
                        LOAD_STR_COL,   load,
-                       LOAD_INT_COL,   MIN(dom->load / dom->info.nrVirtCpu, 100),
+                       LOAD_INT_COL,   MIN(avgload, 100),
                        MEMORY_COL,     mem,
                        FOREGROUND_COL, foreground,
                        WEIGHT_COL,     weight,
@@ -355,6 +362,68 @@ static void domain_update_tree_store(struct vconsole_domain *dom,
 
 /* ------------------------------------------------------------------ */
 
+static void domain_vte_geometry_hints(struct vconsole_domain *dom, GtkWindow *win)
+{
+    GdkWindowHints mask = 0;
+    GdkGeometry geo = {};
+    GtkBorder padding = { 0 };
+
+    geo.width_inc  = vte_terminal_get_char_width(VTE_TERMINAL(dom->vte));
+    geo.height_inc = vte_terminal_get_char_height(VTE_TERMINAL(dom->vte));
+    mask |= GDK_HINT_RESIZE_INC;
+    geo.base_width  = geo.width_inc;
+    geo.base_height = geo.height_inc;
+    mask |= GDK_HINT_BASE_SIZE;
+    geo.min_width  = geo.width_inc * 80;
+    geo.min_height = geo.height_inc * 25;
+    mask |= GDK_HINT_MIN_SIZE;
+
+    gtk_style_context_get_padding(
+        gtk_widget_get_style_context(dom->vte),
+        gtk_widget_get_state_flags(dom->vte),
+        &padding);
+
+    geo.base_width  += padding.left + padding.right;
+    geo.base_height += padding.top + padding.bottom;
+    geo.min_width   += padding.left + padding.right;
+    geo.min_height  += padding.top + padding.bottom;
+
+    gtk_window_set_geometry_hints(win, dom->vte, &geo, mask);
+}
+
+static gboolean domain_window_close(GtkWidget *widget, GdkEvent *event,
+                                    void *opaque)
+{
+    struct vconsole_domain *dom = opaque;
+    struct vconsole_window *win = dom->conn->win;
+    GtkWidget *lhbox;
+
+    gtk_widget_reparent(dom->vbox, win->notebook);
+    lhbox = tab_label_with_close_button(dom->name,
+                                        G_CALLBACK(domain_close_tab_btn),
+                                        dom);
+    gtk_notebook_set_tab_label(GTK_NOTEBOOK(win->notebook),
+                               dom->vbox, lhbox);
+    gtk_widget_destroy(dom->window);
+    dom->window = NULL;
+    return TRUE;
+}
+
+void domain_untabify(struct vconsole_domain *dom)
+{
+    if (dom->window)
+        return;
+
+    dom->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
+    gtk_widget_reparent(dom->vbox, dom->window);
+    gtk_window_set_title(GTK_WINDOW(dom->window), dom->name);
+    domain_vte_geometry_hints(dom, GTK_WINDOW(dom->window));
+    g_signal_connect(dom->window, "delete-event",
+                     G_CALLBACK(domain_window_close), dom);
+
+    gtk_widget_show_all(dom->window);
+}
+
 void domain_start(struct vconsole_domain *dom)
 {
     virDomainPtr d = virDomainLookupByUUIDString(dom->conn->ptr, dom->uuid);
@@ -624,11 +693,39 @@ void domain_update_all(struct vconsole_window *win)
     }
 }
 
+static void domain_close_tab_btn(GtkWidget *btn, gpointer opaque)
+{
+    struct vconsole_domain *dom = opaque;
+    virDomainPtr d = virDomainLookupByUUIDString(dom->conn->ptr, dom->uuid);
+
+    domain_close_tab(dom, d);
+}
+
+static GtkWidget *tab_label_with_close_button(const char *labeltext,
+                                              GCallback callback,
+                                              gpointer opaque)
+{
+    GtkWidget *label, *lclose, *limg, *lhbox;
+
+    label = gtk_label_new(labeltext);
+    lclose = gtk_button_new();
+    g_signal_connect(lclose, "clicked", callback, opaque);
+    limg = gtk_image_new_from_stock(GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
+    gtk_button_set_image(GTK_BUTTON(lclose), limg);
+    gtk_button_set_always_show_image(GTK_BUTTON(lclose), TRUE);
+    lhbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 1);
+    gtk_box_pack_start(GTK_BOX(lhbox), label, TRUE, TRUE, 0);
+    gtk_box_pack_start(GTK_BOX(lhbox), lclose, FALSE, TRUE, 0);
+    gtk_widget_show_all(lhbox);
+
+    return lhbox;
+}
+
 void domain_activate(struct vconsole_domain *dom)
 {
     virDomainPtr d = virDomainLookupByUUIDString(dom->conn->ptr, dom->uuid);
     struct vconsole_window *win = dom->conn->win;
-    GtkWidget *label, *fstatus;
+    GtkWidget *lhbox, *fstatus;
     gint page;
 
     if (dom->vte) {
@@ -642,24 +739,30 @@ void domain_activate(struct vconsole_domain *dom)
         g_signal_connect(dom->vte, "commit",
                          G_CALLBACK(domain_user_input), dom);
         vte_terminal_set_scrollback_lines(VTE_TERMINAL(dom->vte), 9999);
+        vte_terminal_set_size(VTE_TERMINAL(dom->vte), 80, 24);
 
         dom->status = gtk_label_new("-");
         gtk_misc_set_alignment(GTK_MISC(dom->status), 0, 0.5);
         gtk_misc_set_padding(GTK_MISC(dom->status), 3, 1);
 
-        dom->vbox = gtk_vbox_new(FALSE, 1);
+        dom->vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 1);
         gtk_container_set_border_width(GTK_CONTAINER(dom->vbox), 1);
         gtk_box_pack_start(GTK_BOX(dom->vbox), dom->vte, TRUE, TRUE, 0);
         fstatus = gtk_frame_new(NULL);
         gtk_box_pack_end(GTK_BOX(dom->vbox), fstatus, FALSE, TRUE, 0);
         gtk_container_add(GTK_CONTAINER(fstatus), dom->status);
 
-        label = gtk_label_new(dom->name);
+        lhbox = tab_label_with_close_button(dom->name,
+                                            G_CALLBACK(domain_close_tab_btn),
+                                            dom);
         page = gtk_notebook_insert_page(GTK_NOTEBOOK(win->notebook),
-                                        dom->vbox, label, -1);
+                                        dom->vbox, lhbox, -1);
+        gtk_notebook_set_tab_reorderable(GTK_NOTEBOOK(win->notebook),
+                                         dom->vbox, TRUE);
         gtk_widget_show_all(dom->vbox);
         gtk_notebook_set_current_page(GTK_NOTEBOOK(win->notebook), page);
         domain_configure_vte(dom);
+        domain_vte_geometry_hints(dom, GTK_WINDOW(win->toplevel));
 
         domain_update_info(dom, d);
         if (dom->info.state == VIR_DOMAIN_RUNNING)
