@@ -57,10 +57,9 @@ void gtk_message(GtkWidget *parent, GtkWidget **dialog, GtkMessageType type,
     GtkWidget *d = NULL;
     va_list args;
     char msgbuf[1024];
-    int rc;
 
     va_start(args, fmt);
-    rc = vsnprintf(msgbuf, sizeof(msgbuf), fmt, args);
+    vsnprintf(msgbuf, sizeof(msgbuf), fmt, args);
     va_end(args);
     if (debug)
 	fprintf(stderr, "%s: %s", gtk_msg_type_name[type], msgbuf);
@@ -107,15 +106,18 @@ static int gtk_getstring(GtkWidget *window, char *title, char *message,
     dialog = gtk_dialog_new_with_buttons(title,
                                          GTK_WINDOW(window),
                                          GTK_DIALOG_DESTROY_WITH_PARENT,
-                                         GTK_STOCK_OK,
-                                         GTK_RESPONSE_ACCEPT,
-                                         GTK_STOCK_CANCEL,
-                                         GTK_RESPONSE_REJECT,
+                                         "_OK",     GTK_RESPONSE_ACCEPT,
+                                         "_Cancel", GTK_RESPONSE_REJECT,
                                          NULL);
     gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);
 
     label = gtk_label_new(message);
+#if GTK_CHECK_VERSION(3,16,0)
+    gtk_label_set_xalign(GTK_LABEL(label), 0);
+    gtk_label_set_yalign(GTK_LABEL(label), 0.5);
+#else
     gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
+#endif
 
     entry = gtk_entry_new();
     gtk_entry_set_text(GTK_ENTRY(entry), dest);
@@ -186,21 +188,29 @@ static void menu_cb_close_app(GtkAction *action, gpointer userdata)
     gtk_widget_destroy(win->toplevel);
 }
 
+static gboolean terminal_font_filter(const PangoFontFamily *family,
+                                     const PangoFontFace   *face,
+                                     gpointer               data)
+{
+    return pango_font_family_is_monospace((PangoFontFamily *) family);
+}
+
 static void menu_cb_config_font(GtkAction *action, void *data)
 {
     struct vconsole_window *win = data;
     GtkWidget *dialog;
 
-    dialog = gtk_font_selection_dialog_new("Terminal font");
+    dialog = gtk_font_chooser_dialog_new("Terminal font",
+                                         GTK_WINDOW(win->toplevel));
     if (win->tty_font)
-        gtk_font_selection_dialog_set_font_name
-            (GTK_FONT_SELECTION_DIALOG(dialog), win->tty_font);
+        gtk_font_chooser_set_font(GTK_FONT_CHOOSER(dialog), win->tty_font);
+    gtk_font_chooser_set_filter_func(GTK_FONT_CHOOSER(dialog),
+                                     terminal_font_filter, NULL, NULL);
 
     gtk_widget_show_all(dialog);
     switch (gtk_dialog_run(GTK_DIALOG(dialog))) {
     case GTK_RESPONSE_OK:
-        win->tty_font = gtk_font_selection_dialog_get_font_name
-	    (GTK_FONT_SELECTION_DIALOG(dialog));
+        win->tty_font = gtk_font_chooser_get_font(GTK_FONT_CHOOSER(dialog));
 	g_key_file_set_string(config, "tty", "font", win->tty_font);
         config_write();
         domain_configure_all_vtes(win);
@@ -209,27 +219,24 @@ static void menu_cb_config_font(GtkAction *action, void *data)
     gtk_widget_destroy(dialog);
 }
 
-static int pickcolor(char *title, char *group, char *key, char *current)
+static int pickcolor(GtkWindow *parent, char *title,
+                     char *group, char *key, char *current)
 {
     GtkWidget *dialog;
-    GdkColor color = {0,0,0,0};
-    GtkColorSelection *csel;
-    char name[16];
+    GdkRGBA color = { 0, 0, 0, 0 };
+    char *name;
     int rc = -1;
 
-    gdk_color_parse(current, &color);
-    dialog = gtk_color_selection_dialog_new(title);
-    csel = GTK_COLOR_SELECTION(gtk_color_selection_dialog_get_color_selection
-                               (GTK_COLOR_SELECTION_DIALOG(dialog)));
-    gtk_color_selection_set_has_opacity_control(csel, FALSE);
-    gtk_color_selection_set_current_color(csel, &color);
+    gdk_rgba_parse(&color, current);
+    dialog = gtk_color_chooser_dialog_new(title, parent);
+    gtk_color_chooser_set_use_alpha(GTK_COLOR_CHOOSER(dialog), false);
+    gtk_color_chooser_set_rgba(GTK_COLOR_CHOOSER(dialog), &color);
 
     gtk_widget_show_all(dialog);
     switch (gtk_dialog_run(GTK_DIALOG(dialog))) {
     case GTK_RESPONSE_OK:
-	gtk_color_selection_get_current_color(csel, &color);
-        snprintf(name, sizeof(name), "#%04x%04x%04x",
-                 color.red, color.green, color.blue);
+        gtk_color_chooser_get_rgba(GTK_COLOR_CHOOSER(dialog), &color);
+        name = gdk_rgba_to_string(&color);
         g_key_file_set_string(config, group, key, name);
         config_write();
 	rc = 0;
@@ -243,7 +250,8 @@ static void menu_cb_config_fg(GtkAction *action, void *data)
     struct vconsole_window *win = data;
     GError *err = NULL;
 
-    if (0 != pickcolor("Terminal text color", "tty", "foreground", win->tty_fg))
+    if (0 != pickcolor(GTK_WINDOW(win->toplevel), "Terminal text color",
+                       "tty", "foreground", win->tty_fg))
 	return;
     win->tty_fg = g_key_file_get_string(config, "tty", "foreground", &err);
     domain_configure_all_vtes(win);
@@ -254,7 +262,8 @@ static void menu_cb_config_bg(GtkAction *action, void *data)
     struct vconsole_window *win = data;
     GError *err = NULL;
 
-    if (0 != pickcolor("Terminal background", "tty", "background", win->tty_bg))
+    if (0 != pickcolor(GTK_WINDOW(win->toplevel), "Terminal background",
+                       "tty", "background", win->tty_bg))
 	return;
     win->tty_bg = g_key_file_get_string(config, "tty", "background", &err);
     domain_configure_all_vtes(win);
@@ -292,7 +301,6 @@ static struct vconsole_domain *find_guest(struct vconsole_window *win)
 }
 
 static void run_virt_viewer(struct vconsole_domain *dom,
-                            bool direct,
                             bool reconnect)
 {
     char *argv[32];
@@ -303,8 +311,7 @@ static void run_virt_viewer(struct vconsole_domain *dom,
 
     argv[argc++] = "virt-viewer";
     argv[argc++] = "--wait";
-    if (direct)
-        argv[argc++] = "--direct";
+    argv[argc++] = "--attach";
     if (reconnect)
         argv[argc++] = "--reconnect";
     argv[argc++] = "-c";
@@ -325,13 +332,61 @@ static void run_virt_viewer(struct vconsole_domain *dom,
     }
 }
 
+static void run_virsh_edit(struct vconsole_domain *dom)
+{
+    char *argv[32];
+    int argc = 0;
+    char *uri;
+
+    uri = virConnectGetURI(dom->conn->ptr);
+
+    argv[argc++] = "xterm";
+    argv[argc++] = "-e";
+    argv[argc++] = "virsh";
+    argv[argc++] = "-c";
+    argv[argc++] = uri;
+    argv[argc++] = "edit";
+    argv[argc++] = dom->uuid;
+    argv[argc++] = NULL;
+    assert(argc < sizeof(argv)/sizeof(argv[0]));
+
+    if (fork() <= 0) {
+        /* parent */
+        free(uri);
+        return;
+    } else {
+        /* child */
+        execvp("xterm", argv);
+        perror("execvp");
+        exit(1);
+    }
+}
+
+static void menu_cb_untabify(GtkToggleAction *action, gpointer userdata)
+{
+    struct vconsole_window *win = userdata;
+    struct vconsole_domain *dom = find_guest(win);
+
+    if (dom)
+        domain_untabify(dom);
+}
+
+static void menu_cb_vm_edit(GtkAction *action, void *data)
+{
+    struct vconsole_window *win = data;
+    struct vconsole_domain *dom = find_guest(win);
+
+    if (dom)
+        run_virsh_edit(dom);
+}
+
 static void menu_cb_vm_gfx(GtkAction *action, void *data)
 {
     struct vconsole_window *win = data;
     struct vconsole_domain *dom = find_guest(win);
 
     if (dom)
-        run_virt_viewer(dom, true, true);
+        run_virt_viewer(dom, true);
 }
 
 static void menu_cb_vm_run(GtkAction *action, void *data)
@@ -349,7 +404,7 @@ static void menu_cb_vm_run_gfx(GtkAction *action, void *data)
     struct vconsole_domain *dom = find_guest(win);
 
     if (dom) {
-        run_virt_viewer(dom, true, false);
+        run_virt_viewer(dom, false);
         domain_start(dom);
     }
 }
@@ -419,7 +474,6 @@ static void menu_cb_about(GtkAction *action, gpointer userdata)
                           "authors",         authors,
                           "comments",        comments,
                           "copyright",       copyright,
-                          "logo-icon-name",  GTK_STOCK_ABOUT,
                           "version",         VERSION,
                           NULL);
 }
@@ -515,8 +569,16 @@ static const GtkActionEntry entries[] = {
 	.label       = "Terminal _background ...",
 	.callback    = G_CALLBACK(menu_cb_config_bg),
     },{
+	.name        = "Untabify",
+	.label       = "_Detach Tab",
+	.callback    = G_CALLBACK(menu_cb_untabify),
+    },{
 
         /* --- guest menu --- */
+	.name        = "GuestEdit",
+	.label       = "Change guest configuration",
+	.callback    = G_CALLBACK(menu_cb_vm_edit),
+    },{
 	.name        = "GuestGfx",
 	.label       = "Show graphic console",
 	.callback    = G_CALLBACK(menu_cb_vm_gfx),
@@ -528,7 +590,9 @@ static const GtkActionEntry entries[] = {
 	.callback    = G_CALLBACK(menu_cb_vm_run),
     },{
 	.name        = "GuestRunGfx",
-	.label       = "Run and show grapics console",
+	.stock_id    = "window-new",
+	.label       = "Run with graphic console",
+	.tooltip     = "Run guest and show graphic console",
 	.callback    = G_CALLBACK(menu_cb_vm_run_gfx),
     },{
 	.name        = "GuestPause",
@@ -545,25 +609,25 @@ static const GtkActionEntry entries[] = {
 	.callback    = G_CALLBACK(menu_cb_vm_save),
     },{
 	.name        = "GuestReboot",
+	.stock_id    = GTK_STOCK_REFRESH,
 	.label       = "Reboot",
         .tooltip     = "Reboot guest",
 	.callback    = G_CALLBACK(menu_cb_vm_reboot),
     },{
 	.name        = "GuestShutdown",
+	.stock_id    = GTK_STOCK_MEDIA_STOP,
 	.label       = "Shutdown",
         .tooltip     = "Shutdown guest",
 	.callback    = G_CALLBACK(menu_cb_vm_shutdown),
     },{
 	.name        = "GuestReset",
-	.stock_id    = GTK_STOCK_REFRESH,
 	.label       = "Reset",
         .tooltip     = "Reset guest",
 	.callback    = G_CALLBACK(menu_cb_vm_reset),
     },{
 	.name        = "GuestKill",
-	.stock_id    = GTK_STOCK_MEDIA_STOP,
 	.label       = "Destroy",
-        .tooltip     = "Destriy guest",
+        .tooltip     = "Destroy guest",
 	.callback    = G_CALLBACK(menu_cb_vm_kill),
 
     },{
@@ -616,10 +680,14 @@ static char ui_xml[] =
 "      <menuitem action='TerminalBlink'/>\n"
 "      <separator/>\n"
 "      <menuitem action='FullScreen'/>\n"
+"      <separator/>\n"
+"      <menuitem action='Untabify'/>\n"
 "    </menu>\n"
 "    <menu action='GuestMenu'>\n"
 "      <menuitem action='GuestLogging'/>\n"
 "      <separator/>\n"
+"      <menuitem action='GuestEdit'/>\n"
+"      <separator/>\n"
 "      <menuitem action='GuestGfx'/>\n"
 "      <separator/>\n"
 "      <menuitem action='GuestRun'/>\n"
@@ -637,16 +705,13 @@ static char ui_xml[] =
 "    </menu>\n"
 "  </menubar>\n"
 "  <toolbar action='ToolBar'>"
-"    <toolitem action='CloseTab'/>\n"
+"    <toolitem action='GuestRunGfx'/>\n"
 "    <separator/>\n"
 "    <toolitem action='GuestRun'/>\n"
 "    <toolitem action='GuestPause'/>\n"
 "    <toolitem action='GuestSave'/>\n"
-"    <separator/>\n"
-"    <toolitem action='GuestReset'/>\n"
-"    <toolitem action='GuestKill'/>\n"
-"    <separator/>\n"
-"    <toolitem action='GuestRunGfx'/>\n"
+"    <toolitem action='GuestReboot'/>\n"
+"    <toolitem action='GuestShutdown'/>\n"
 "  </toolbar>\n"
 "</ui>\n";
 
@@ -753,7 +818,7 @@ static struct vconsole_window *vconsole_toplevel_create(void)
     win = g_new0(struct vconsole_window, 1);
     win->toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
     gtk_window_set_title(GTK_WINDOW(win->toplevel), APPNAME);
-    gtk_window_set_default_size(GTK_WINDOW(win->toplevel), 640, 400);
+    gtk_window_set_default_size(GTK_WINDOW(win->toplevel), 800, 600);
     g_signal_connect(G_OBJECT(win->toplevel), "destroy",
 		     G_CALLBACK(destroy), win);
     g_signal_connect(G_OBJECT(win->toplevel), "window-state-event",
@@ -778,9 +843,13 @@ static struct vconsole_window *vconsole_toplevel_create(void)
 
     /* main area */
     win->notebook = gtk_notebook_new();
+    gtk_notebook_set_tab_pos(GTK_NOTEBOOK(win->notebook), GTK_POS_TOP);
+    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(win->notebook), TRUE);
+    gtk_notebook_set_scrollable(GTK_NOTEBOOK(win->notebook), TRUE);
+    gtk_notebook_popup_enable(GTK_NOTEBOOK(win->notebook));
 
     /* Make a vbox and put stuff in */
-    vbox = gtk_vbox_new(FALSE, 1);
+    vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 1);
     gtk_container_set_border_width(GTK_CONTAINER(vbox), 1);
     gtk_container_add(GTK_CONTAINER(win->toplevel), vbox);
     menubar = gtk_ui_manager_get_widget(win->ui, "/MainMenu");
@@ -925,6 +994,7 @@ static void vconsole_tab_list_create(struct vconsole_window *win)
     /* id */
     renderer = gtk_cell_renderer_text_new();
     g_object_set(renderer, "xalign", 0.5, NULL);
+    g_object_set(renderer, "height", 1, NULL);
     column = gtk_tree_view_column_new_with_attributes("ID",
                                                       renderer,
                                                       "text", ID_COL,
@@ -1048,7 +1118,6 @@ main(int argc, char *argv[])
         uri = getenv("VIRSH_DEFAULT_CONNECT_URI");
 
     /* init */
-    g_thread_init(NULL);
     gvir_event_register();
     config_read();
 
